import numpy as np


def numerical_gradient(f, x: np.ndarray):
    """
    calc gradient
    for now, only support ndim 2
    """
    h = 1e-4
    grad = np.zeros_like(x)
    for i in range(x.size):
        tmp_val = x.flat[i]
        x.flat[i] = tmp_val + h
        fxh1 = f(x)

        x.flat[i] = tmp_val - h
        fxh2 = f(x)
        grad.flat[i] = (fxh1 - fxh2) / (2 * h)
        x.flat[i] = tmp_val
    return grad


def gradient_decline(f, init_x, learn_gap, step=10000):
    """梯度下降，可能单词用错了"""
    x = init_x
    for i in range(step):
        grad = numerical_gradient(f, x)
        x = x - grad * learn_gap
    return f(x)



from sdl.layers import Relu, Affine
import numpy as np


def test_Relu():
    relu = Relu()
    x = np.array([-1, 0, 1, 3])
    y = relu.forward(x)
    assert(np.array_equal(y, [0, 0, 1, 3]))

    back = relu.backward(np.array([-1, 0, 1, 3]))
    assert(np.array_equal(back, [0, 0, 1, 3]))

def test_Affine():
    affine = Affine(np.array([[1, 2], [1, 2], [1, 2]]), np.array([1, 1]))
    y = affine.forward(np.array([1, 2, 3]))
    assert(np.array_equal(y, [7, 13]))

# test_Relu()

test_Affine()



from sdl_x.lose import mean_squared_error
from sdl_x.lose import cross_entropy_error

from sdl_x.gradient import numerical_gradient

from sdl_x.active import softmax
from sdl_x.active import sigmoid

from sdl_x.net import SimpleNet, TwoLayerNet

from sdl_x.mnist import load_mnist

from .layer import Affine
from .layer import Relu
from .layer import SoftmaxWithLoss

import numpy as np


def mean(src: np.ndarray, axis=None) -> np.ndarray:
    """
    求均值
    """
    if axis is None:
        axis = src


import numpy as np


class SGD:
    def __init__(self, lr=0.01):
        self.lr = lr

    def update(self, params, grads):
        for key in params.keys():
            params[key] -= self.lr * grads[key]


class Momentum:
    def __init__(self, lr=0.01, momentum=0.9):
        self.lr = lr
        self.momentum = momentum
        self.v = None

    def update(self, params, grads):
        if self.v in None:
            self.v = {}
            for key, val in params.iterms():
                self.v[key] = np.zeros_like(val)

        for key in params.keys():
            self.v[key] = self.momentum * self.v[key] - self.lr * grads[key]
            params[key] += self.v[key]


class AddGrad:
    def __init__(self, lr=0.01):
        self.lr = lr
        self.h = None

    def update(self, params, grads):
        if self.h is None:
            self.h = {}
            for key, val in params.iterms():
                self.h[key] = np.zeros_like(val)

        for key in params.keys():
            self.h[key] += grads[key] * grads[key]
            params[key] -= self.lr * grads[key] / (np.sqrt(self.h[key] + 1e-7))


class Adam:
    """Adam (http://arxiv.org/abs/1412.6980v8)"""

    def __init__(self, lr=0.001, beta1=0.9, beta2=0.999):
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.iter = 0
        self.m = None
        self.v = None

    def update(self, params, grads):
        if self.m is None:
            self.m, self.v = {}, {}
            for key, val in params.items():
                self.m[key] = np.zeros_like(val)
                self.v[key] = np.zeros_like(val)

        self.iter += 1
        lr_t = self.lr * np.sqrt(1.0 - self.beta2 ** self.iter) / (1.0 - self.beta1 ** self.iter)

        for key in params.keys():
            self.m[key] += (1 - self.beta1) * (grads[key] - self.m[key])
            self.v[key] += (1 - self.beta2) * (grads[key] ** 2 - self.v[key])
            params[key] -= lr_t * self.m[key] / (np.sqrt(self.v[key]) + 1e-7)



import numpy as np
from sdl_x import softmax
from sdl_x import cross_entropy_error
from sdl_x import numerical_gradient
from .layer import Affine
from .layer import Relu
from .layer import SoftmaxWithLoss
from collections import OrderedDict
from sdl_x import active


class SimpleNet:
    def __init__(self):
        self.W = np.random.randn(2, 3)

    def predict(self, x):
        return np.dot(x, self.W)

    def loss(self, x, t):
        z = self.predict(x)
        y = softmax(z)
        loss = cross_entropy_error(y, t)

        return loss


class MultiLayerNet:
    def __init__(self, input_size, hidden_size, output_size, weight_init_std=0.01):
        """多层网络
        这里，我们写死激活函数和输出函数以及lose函数。在需要的时候可以改动
        没有进行测试
        @:param
        hidden_size 为一个数组，中间层的输出个数
        """
        self.input_size = input_size
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.weight_init_std = weight_init_std
        self.params = {}
        self.layers = OrderedDict()
        self._loop_net(self._create_net)
        self.lastLayer = SoftmaxWithLoss()

    def _loop_net(self, func):
        """除了最后的激活函数（输出函数），遍历所有的层
        func: 遍历用到的函数。参数：第几层，输入大小，输出大小，是否是最后的激活层"""
        index = 0
        input_size = self.input_size
        for index in range(0, len(self.hidden_size)):
            output_size = self.hidden_size[index]
            func(index, input_size, output_size, False)
            input_size = output_size
        func(index + 1, input_size, self.output_size, True)

    def _create_net(self, index, input_size, output_size, is_last):
        W_index = 'W{}'.format(index)
        b_index = 'b{}'.format(index)
        self.params[W_index] = self.weight_init_std * np.random.randn(input_size, output_size)
        self.params[b_index] = np.zeros(output_size)
        self.layers['Affine{}'.format(index)] = Affine(self.params[W_index], self.params[b_index])
        if not is_last:
            self.layers['Relu{}'.format(index)] = Relu()

    def predict(self, x):
        for layer in self.layers.values():
            x = layer.forward(x)
        return x

    def loss(self, x, t):
        y = self.predict(x)
        return self.lastLayer.forward(y, t)

    def accuracy(self, x, t):
        y = self.predict(x)
        y = np.argmax(y, axis=1)
        if t.ndim != 1:
            t = np.argmax(t, axis=1)
        accuracy = np.sum(y == t) / float(x.shape[0])
        return accuracy

    def _calc_numerical_gradient(self, index, grads, x, t):
        loss_func = lambda W: self.loss(x, t)
        W_index = 'W{}'.format(index)
        b_index = 'b{}'.format(index)

        grads[W_index] = numerical_gradient(loss_func, self.params[W_index])
        grads[b_index] = numerical_gradient(loss_func, self.params[b_index])

    def numerical_gradient(self, x, t):
        grads = {}
        self._loop_net(lambda index, input_size, output_size, is_last: self._calc_numerical_gradient(index, grads, x, t))
        return grads

    def _calc_gradient(self, index, grads):
        grads['W{}'.format(index)] = self.layers['Affine{}'.format(index)].dW
        grads['b{}'.format(index)] = self.layers['Affine{}'.format(index)].db

    def gradient(self, x, t):
        # forward
        self.loss(x, t)

        # backward
        dout = 1
        dout = self.lastLayer.backward(dout)

        layers = list(self.layers.values())
        layers.reverse()
        for layer in layers:
            dout = layer.backward(dout)

        grads = {}
        self._loop_net(lambda index, input_size, output_size, is_last: \
                           self._calc_gradient(index, grads))

        return grads


class TwoLayerNet(MultiLayerNet):
    def __init__(self, input_size, hidden_size, output_size, weight_init_std=0.01):
        MultiLayerNet.__init__(self, input_size, [hidden_size], output_size, weight_init_std)


class TwoLayerNet_OLD:
    def __init__(self, input_size, hidden_size, output_size, weight_init_std=0.01):
        self.params = {'W1': weight_init_std * \
                             np.random.randn(input_size, hidden_size), 'b1': np.zeros(hidden_size),
                       'W2': weight_init_std * \
                             np.random.randn(hidden_size, output_size), 'b2': np.zeros(output_size)}

        # 生成层
        self.layers = OrderedDict()
        self.layers['Affine1'] = Affine(self.params['W1'], self.params['b1'])
        self.layers['Relu1'] = Relu()
        self.layers['Affine2'] = Affine(self.params['W2'], self.params['b2'])

        self.lastLayer = SoftmaxWithLoss()

    def predict(self, x):
        for layer in self.layers.values():
            x = layer.forward(x)
        return x

    def loss(self, x, t):
        y = self.predict(x)
        return self.lastLayer.forward(y, t)

    def accuracy(self, x, t):
        y = self.predict(x)
        y = np.argmax(y, axis=1)
        if t.ndim != 1:
            t = np.argmax(t, axis=1)
        accuracy = np.sum(y == t) / float(x.shape[0])
        return accuracy

    def numerical_gradient(self, x, t):
        loss_func = lambda W: self.loss(x, t)

        grads = {'W1': numerical_gradient(loss_func, self.params['W1']),
                 'b1': numerical_gradient(loss_func, self.params['b1']),
                 'W2': numerical_gradient(loss_func, self.params['W2']),
                 'b2': numerical_gradient(loss_func, self.params['b2'])}
        return grads

    def gradient(self, x, t):
        # forward
        self.loss(x, t)

        # backward
        dout = 1
        dout = self.lastLayer.backward(dout)

        layers = list(self.layers.values())
        layers.reverse()
        for layer in layers:
            dout = layer.backward(dout)

        grads = {'W1': self.layers['Affine1'].dW,
                 'b1': self.layers['Affine1'].db,
                 'W2': self.layers['Affine2'].dW,
                 'b2': self.layers['Affine2'].db}
        return grads



class SimpleTwoLayerNet():
    """只用到基本概念的net"""

    def __init__(self, input_size, hidden_size, output_size, weight_init_std=0.01):
        self.params = {'W1': weight_init_std * \
                             np.random.randn(input_size, hidden_size),
                       'b1': np.zeros(hidden_size),
                       'W2': weight_init_std * \
                             np.random.randn(hidden_size, output_size),
                       'b2': np.zeros(output_size)}

    def predict(self, x):
        W1, W2 = self.params['W1'], self.params['W2']
        b1, b2 = self.params['b1'], self.params['b2']

        a1 = np.dot(x, W1) + b1
        z1 = active.sigmoid(a1)
        a2 = np.dot(z1, W2) + b2
        y = active.softmax(a2)

        return y

    def loss(self, x, t):
        y = self.predict(x)
        return cross_entropy_error(y, t)

    def accuracy(self, x, t):
        y = self.predict(x)
        y = np.argmax(y, axis=1)
        if t.ndim != 1:
            t = np.argmax(t, axis=1)
        accuracy = np.sum(y == t) / float(x.shape[0])
        return accuracy

    def numerical_gradient(self, x, t):
        loss_func = lambda W: self.loss(x, t)

        grads = {'W1': numerical_gradient(loss_func, self.params['W1']),
                 'b1': numerical_gradient(loss_func, self.params['b1']),
                 'W2': numerical_gradient(loss_func, self.params['W2']),
                 'b2': numerical_gradient(loss_func, self.params['b2'])}
        return grads



import numpy as np
from sdl_x import softmax
from sdl_x import cross_entropy_error


class Layer:
    def backward(self, dout):
        pass


class MulLayer(Layer):
    def __init__(self):
        self.x = None
        self.y = None

    def forward(self, x, y):
        self.x = x
        self.y = y
        return x * y

    def backward(self, dout):
        return dout * self.y, dout * self.x


class AddLayer:
    def __init__(self):
        pass

    def forward(self, x, y):
        return x + y

    def backward(self, dout):
        return dout, dout


class Relu:
    def __init__(self):
        self.mask = None

    def forward(self, x):
        self.mask = (x <= 0)
        out = x.copy()
        out[self.mask] = 0
        return out

    def backward(self, dout):
        rst = dout.copy()
        rst[self.mask] = 0
        return rst


class Sigmoid(Layer):
    def __init__(self):
        self.out = None

    def forward(self, x):
        self.out = 1 / (1 + np.exp(-x))
        return self.out

    def backward(self, dout):
        return dout * self.out * (1 - self.out)


class Affine:
    def __init__(self, W, b):
        self.W = W
        self.b = b
        self.x = None
        self.dW = None
        self.db = None

    def forward(self, x):
        self.x = x
        out = np.dot(x, self.W) + self.b
        return out

    def backward(self, dout):
        dx = np.dot(dout, self.W.T)
        self.dW = np.dot(self.x.T, dout)
        self.db = np.sum(dout, axis=0)

        return dx


class SoftmaxWithLoss:
    def __init__(self):
        self.loss = None
        self.y = None
        self.t = None

    def forward(self, x, t):
        self.t = t
        self.y = softmax(x)
        self.loss = cross_entropy_error(self.y, self.t)

        return self.loss

    def backward(self, dout = 1):
        """dout按理说为实数"""
        batch_size = self.t.shape[0]
        dx = (self.y - self.t) / batch_size * dout
        return dx





# softmax函数
import numpy as np


def relu(x):
    return np.maximum(0, x)


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def softmax(x):
    """
    y[k] = exp(e[k]) / Sum(exp[1~n]))
    """
    # 换成转置，省的处理(3,)这种shape
    x = x.T
    x = x - x.max(axis=0)
    exp = np.exp(x)
    y = exp / exp.sum(axis=0)
    return y.T


import numpy as np
# this defines some calculate loss function


# 均方误差
def mean_squared_error(y, t):
    """Sum((y[k] - t[k]) ** 2) k = 1~n"""
    return 0.5 * np.sum((y - t) * 1)


# 交叉熵误差
# 监督数据是one-hot的形式
def cross_entropy_error_one_shot(y, t):
    delta = 1e-7
    if y.ndim == 1:
        y = y.reshape(1, y.size)
        t = t.reshape(1, t.size)

    batch_size = y.shape[0]
    return -np.sum(t * np.log(y + delta)) / batch_size


# 监督数据是标签形式
def cross_entropy_error_label(y, t):
    delta = 1e-7
    if y.ndim == 1:
        y = y.reshape(1, y.size)
        t = t.reshape(1, t.size)

    batch_size = y.shape[0]
    return -np.sum(np.log(y[np.arange(batch_size), t] + 1e-7)) / batch_size


# entropy 翻译为熵
def cross_entropy_error(y, t, one_hot_label=True):
    """-Sum(t[k]log(y[k]))
    why you so complex
    """
    # 保护log(0)情况
    delta = 1e-7

    size = y.shape[0]
    if one_hot_label:
        return -np.sum(np.log(y + delta) * t) / size
    else:
        return -np.sum(np.log(y[np.arange(size), t] + delta)) / size


# coding: utf-8
try:
    import urllib.request
except ImportError:
    raise ImportError('You should use Python 3.x')
import os.path
import gzip
import pickle
import os
import numpy as np


url_base = 'http://yann.lecun.com/exdb/mnist/'
key_file = {
    'train_img':'train-images-idx3-ubyte.gz',
    'train_label':'train-labels-idx1-ubyte.gz',
    'test_img':'t10k-images-idx3-ubyte.gz',
    'test_label':'t10k-labels-idx1-ubyte.gz'
}

dataset_dir = os.path.abspath(os.getcwd()) + "/cache"
save_file = dataset_dir + "/mnist.pkl"

train_num = 60000
test_num = 10000
img_dim = (1, 28, 28)
img_size = 784


def _ensure_dataset_dir():
    if not os.path.exists(dataset_dir):
        os.makedirs(dataset_dir)


def _download(file_name):
    _ensure_dataset_dir()
    file_path = dataset_dir + "/" + file_name
    
    if os.path.exists(file_path):
        return

    print("Downloading " + file_name + " ... ")
    urllib.request.urlretrieve(url_base + file_name, file_path)
    print("Done")


def download_mnist():
    for v in key_file.values():
       _download(v)


def _load_label(file_name):
    file_path = dataset_dir + "/" + file_name
    
    print("Converting " + file_name + " to NumPy Array ...")
    with gzip.open(file_path, 'rb') as f:
            labels = np.frombuffer(f.read(), np.uint8, offset=8)
    print("Done")
    
    return labels


def _load_img(file_name):
    file_path = dataset_dir + "/" + file_name
    
    print("Converting " + file_name + " to NumPy Array ...")    
    with gzip.open(file_path, 'rb') as f:
            data = np.frombuffer(f.read(), np.uint8, offset=16)
    data = data.reshape(-1, img_size)
    print("Done")
    
    return data


def _convert_numpy():
    dataset = {'train_img': _load_img(key_file['train_img']),
               'train_label': _load_label(key_file['train_label']),
               'test_img': _load_img(key_file['test_img']),
               'test_label': _load_label(key_file['test_label'])}

    return dataset


def init_mnist():
    download_mnist()
    dataset = _convert_numpy()
    print("Creating pickle file ...")
    with open(save_file, 'wb') as f:
        pickle.dump(dataset, f, -1)
    print("Done!")


def _change_one_hot_label(X):
    """
    convert [0, 8, 1, 1]
    to [[1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
        [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0, 0, 0]]
    """
    T = np.zeros((X.size, 10))
    for idx, row in enumerate(T):
        row[X[idx]] = 1
        
    return T
    

def load_mnist(normalize=True, flatten=True, one_hot_label=False):
    """读入MNIST数据集
    
    Parameters
    ----------
    normalize : 将图像的像素值正规化为0.0~1.0
    flatten : 是否将图像展开为一维数组
    one_hot_label :
        one_hot_label为True的情况下，标签作为one-hot数组返回
        one-hot数组是指[0,0,1,0,0,0,0,0,0,0]这样的数组

    Returns
    -------
    (训练图像, 训练标签), (测试图像, 测试标签)
    """
    if not os.path.exists(save_file):
        init_mnist()
        
    with open(save_file, 'rb') as f:
        dataset = pickle.load(f)
    
    if normalize:
        for key in ('train_img', 'test_img'):
            dataset[key] = dataset[key].astype(np.float32)
            dataset[key] /= 255.0
            
    if one_hot_label:
        dataset['train_label'] = _change_one_hot_label(dataset['train_label'])
        dataset['test_label'] = _change_one_hot_label(dataset['test_label'])
    
    if not flatten:
        for key in ('train_img', 'test_img'):
            dataset[key] = dataset[key].reshape(-1, 1, 28, 28)

    return (dataset['train_img'], dataset['train_label']), (dataset['test_img'], dataset['test_label']) 


if __name__ == '__main__':
    init_mnist()



def getConvolutinSize(h, w, fh, fw, p, s):
    """
    对于填充和步 幅，如何计算输出大小
    :param h: 输入大小为 (H, W)
    :param w:
    :param fh: 滤波器大小为 (FH, FW)
    :param fw:
    :param p: 填充为 P
    :param s: 步幅为 S
    :return: 输出大小
    """
    oh = (h + 2 * p -fh) / s + 1
    oh = (w + 2 * p -fw) / s + 1



from setuptools import setup

setup(
    name='sdl-x',
    version='0.0.2',
    description='collected simple deep learning functions',
    author="huhuang03",
    packages=['sdl_x'],
    install_requires=['numpy']
)
